/*
 * Plasma applet to display application window menus
 *
 * Copyright 2010 Canonical Ltd.
 *
 * Authors:
 * - Aurélien Gâteau <aurelien.gateau@canonical.com>
 *
 * License: GPL v3
 */
// Self
#include "menubarapplet.h"

// Qt
#include <QApplication>
#include <QDBusConnection>
#include <QKeyEvent>
#include <QMenu>
#include <QMouseEvent>
#include <QToolButton>
#include <QVBoxLayout>

// KDE
#include <KConfigDialog>
#include <Plasma/Containment>
#include <Plasma/Corona>
#include <Plasma/IconWidget>

// Local
#include "windowmenumanager.h"
#include "registraradaptor.h"
#include "registrar.h"

static const char* DBUS_SERVICE = "org.ayatana.WindowMenu.Registrar";
static const char* DBUS_OBJECT_PATH = "/org/ayatana/WindowMenu/Registrar";


K_EXPORT_PLASMA_APPLET(menubarapplet, MenuBarApplet)


MenuBarApplet::MenuBarApplet(QObject* parent, const QVariantList& args)
: Plasma::Applet(parent, args)
, mRegistrar(new Registrar(this))
, mLayout(new QGraphicsLinearLayout(this))
, mMouseChecker(new QTimer(this))
, mEmptyMenu(new QMenu)
, mWindowMenuManager(new WindowMenuManager(this))
, mWindowMenu(new QMenu)
, mCurrentButton(0)
{
    setBackgroundHints(StandardBackground);
    setAspectRatioMode(Plasma::IgnoreAspectRatio);
}

MenuBarApplet::~MenuBarApplet()
{
    QDBusConnection::sessionBus().unregisterService(DBUS_SERVICE);
    mEmptyMenu->deleteLater();
    mWindowMenu->deleteLater();
}

void MenuBarApplet::init()
{
    setupWindowMenu();
    setupEmptyMenu();

    mLayout->setContentsMargins(0, 0, 0, 0);
    updateSizePolicy();

    bool ok = QDBusConnection::sessionBus().registerService(DBUS_SERVICE);
    if (!ok) {
        kWarning() << "Could not register" << DBUS_SERVICE;
        return;
    }
    new RegistrarAdaptor(mRegistrar);
    QDBusConnection::sessionBus().registerObject(DBUS_OBJECT_PATH, mRegistrar);
    connect(mRegistrar, SIGNAL(currentMenuChanged()), SLOT(updateMenuBar()));

    connect(mMouseChecker, SIGNAL(timeout()), SLOT(checkMousePosition()));

    connect(this, SIGNAL(activate()), SLOT(slotActivated()));
}

void MenuBarApplet::setupWindowMenu()
{
    QAction* action = mWindowMenu->addAction(i18n("Window"));
    action->setMenu(mWindowMenuManager->menu());
}

void MenuBarApplet::setupEmptyMenu()
{
    QMenu* subMenu = mEmptyMenu->addMenu(i18n("Menu"));
    QAction* action = subMenu->addAction(i18n("No Items"));
    action->setEnabled(false);
}

void MenuBarApplet::updateMenuBar()
{
    qDeleteAll(mMenuButtonList);
    mMenuButtonList.clear();
    mCurrentButton = 0;
    QMenu* menu = mRegistrar->currentMenu();

    if (useButtonFormFactor()) {
        if (menu) {
            createButtonsForButtonFormFactor(menu);
        } else {
            createButtonsForBarFormFactor(mEmptyMenu);
        }
    } else {
        if (!menu) {
            WId winId = mRegistrar->activeWinId();
            if (winId) {
                mWindowMenuManager->setWinId(winId);
                menu = mWindowMenu;
            } else {
                menu = mEmptyMenu;
            }
        }
        createButtonsForBarFormFactor(menu);
    }
}

void MenuBarApplet::slotActivated()
{
    QGraphicsLayoutItem* layoutItem = mLayout->itemAt(0);
    if (!layoutItem) {
        kWarning() << "No layoutItem found!";
        return;
    }
    MenuButton* button = qobject_cast<MenuButton*>(layoutItem->graphicsItem()->toGraphicsObject());
    if (!button) {
        kWarning() << "No button found!";
        return;
    }

    button->nativeWidget()->animateClick();
}

void MenuBarApplet::createButton(const QString& text, QMenu* menu)
{
    MenuButton* button = new MenuButton(this, menu);
    button->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);
    connect(button, SIGNAL(clicked()), SLOT(slotButtonClicked()));
    mLayout->addItem(button);

    button->setText(text);

    disconnect(menu, 0, this, 0); // Ensure we are only connected to aboutToHide() once
    connect(menu, SIGNAL(aboutToHide()), SLOT(slotAboutToHideMenu()));
    mMenuButtonList << button;
}

void MenuBarApplet::createButtonsForButtonFormFactor(QMenu* menu)
{
    createButton(i18n("Menu"), menu);
}

void MenuBarApplet::createButtonsForBarFormFactor(QMenu* menu)
{
    Q_FOREACH(QAction* action, menu->actions()) {
        if (action->isSeparator()) {
            continue;
        }
        if (!action->menu()) {
            kWarning() << "No menu in action" << action->text();
            continue;
        }
        createButton(action->text(), action->menu());
        action->menu()->installEventFilter(this);
    }
}

void MenuBarApplet::slotButtonClicked()
{
    MenuButton* button = qobject_cast<MenuButton*>(sender());
    if (!button) {
        kWarning() << "Not called by a MenuButton!";
        return;
    }
    showMenu(button);
}

void MenuBarApplet::slotAboutToHideMenu()
{
    if (mCurrentButton->menu() == sender()) {
        mCurrentButton->nativeWidget()->setDown(false);
        mCurrentButton = 0;
        mMouseChecker->stop();
    } else {
        kWarning() << "Not called from menu associated with mCurrentButton!";
    }
}

void MenuBarApplet::showMenu(MenuButton* button)
{
    if (mCurrentButton) {
        bool justHide = mCurrentButton == button;
        mCurrentButton->menu()->hide();
        if (justHide) {
            return;
        }
    }
    QMenu* menu = button->menu();
    QPoint pos = containment()->corona()->popupPosition(button, menu->sizeHint());

    mCurrentButton = button;
    mCurrentButton->nativeWidget()->setDown(true);
    // Process gui events so that the first opened button gets some time to
    // repaint itself. If we don't its highlight border stays there while
    // scrubling.
    qApp->processEvents();
    menu->popup(pos);

    mMouseChecker->start();
}

void MenuBarApplet::checkMousePosition()
{
    QPoint pos = view()->mapFromGlobal(QCursor::pos());
    QGraphicsItem* item = view()->itemAt(pos);
    if (!item) {
        return;
    }

    MenuButton* buttonBelow = qobject_cast<MenuButton*>(item->toGraphicsObject());
    if (!buttonBelow) {
        return;
    }

    if (buttonBelow != mCurrentButton) {
        showMenu(buttonBelow);
    }
}

void MenuBarApplet::createConfigurationInterface(KConfigDialog* parent)
{
    QWidget* widget = new QWidget;
    parent->addPage(widget, i18n("General"), icon());
    connect(parent, SIGNAL(applyClicked()), SLOT(applyConfig()));
    connect(parent, SIGNAL(okClicked()), SLOT(applyConfig()));

    mConfigButtonFormFactor = new QCheckBox;
    mConfigButtonFormFactor->setText(i18n("Use Button Form Factor"));
    QVBoxLayout* layout = new QVBoxLayout(widget);
    layout->addWidget(mConfigButtonFormFactor);
    layout->addStretch();

    KConfigGroup cg = config();
    mConfigButtonFormFactor->setChecked(useButtonFormFactor());
}

void MenuBarApplet::applyConfig()
{
    KConfigGroup cg = config();
    cg.writeEntry("useButtonFormFactor", mConfigButtonFormFactor->isChecked());
    updateSizePolicy();
    updateMenuBar();
}

void MenuBarApplet::updateSizePolicy()
{
    if (useButtonFormFactor()) {
        setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);
    } else {
        setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
        setMaximumSize(INT_MAX, INT_MAX);
    }
}

bool MenuBarApplet::useButtonFormFactor() const
{
    return config().readEntry("useButtonFormFactor", false);
}

void MenuBarApplet::showNextPrevMenu(bool next)
{
    int index = mMenuButtonList.indexOf(mCurrentButton);
    if (index == -1) {
        kWarning() << "Couldn't find button!";
        return;
    }
    if (next) {
        index = (index + 1) % mMenuButtonList.count();
    } else {
        index = (index == 0 ? mMenuButtonList.count() : index) - 1;
    }
    showMenu(mMenuButtonList.at(index));
}

bool MenuBarApplet::eventFilter(QObject* object, QEvent* event)
{
    if (event->type() == QEvent::KeyPress) {
        object->removeEventFilter(this);
        QApplication::sendEvent(object, event);
        object->installEventFilter(this);
        if (!event->isAccepted()) {
            QKeyEvent* keyEvent = static_cast<QKeyEvent*>(event);
            switch (keyEvent->key()) {
            case Qt::Key_Left:
                showNextPrevMenu(false);
                break;
            case Qt::Key_Right:
                showNextPrevMenu(true);
                break;
            default:
                break;
            }
        }
        return true;
    }
    return false;
}

#include "menubarapplet.moc"
